//Chris Efstathiou February 2021

#include <ESP8266WiFi.h>
#include <WiFiUdp.h>

//////////  setup definitions  ///////////////

//The paparazzi running laptop connects directly to ESP8266
#define MODE_AP         1  // 0= not used, 1 = active
// ESP8266 connects to the WiFi router where the paparazzi running laptop is also connected.
#define MODE_STA        0 // 0= not used, 1 = active

#define UART_BAUD       57600
#define PACKET_TIMEOUT  5 // After that inactivity time in ms send packet.
#define BUFFER_SIZE     256

#define DEBUG_MODE      0 // if set to 1 all the incoming packets and relevant info are fowarded to the USB serial port.

///////////////   global variables  /////////////

WiFiUDP udp;
IPAddress remote_ip(192, 168, 4, 100); // usual address given from this server to laptop running paparazzi
unsigned int remote_port = 4242; // the paparazzi link agent listens to this port and replies on port 4243.

uint8_t uart_buffer[BUFFER_SIZE];
uint16_t uart_idx = 0;

uint8_t udp_buffer[BUFFER_SIZE];
uint16_t udp_idx = 0;


#if defined(MODE_AP) && MODE_AP == 1 // TESTED, IT WORKS FINE...
const char *ssid = "PAPARAZZI";  // Connect the paparazzi running laptop on this AP
const char *pw = "paparazzi"; // AP password
IPAddress ip(192, 168, 4, 1); // SERVER Ip address
IPAddress netmask(255, 255, 255, 0); // AP Subnet mask.
const int port = 4243; // Paparazzi link agent sends UDP packets to this address
// You must connect the paparazzi data link agent to this ip.
#endif

#if defined(MODE_STA) && MODE_STA == 1
const char *ssid = "router_ssid";  // your router SSID
const char *pw = "password"; // Your router's WiFi PASSWORD
const int port = 4243; // Paparazzi link agent sends UDP packets to this address
// Not tested but after connecting the paparazzi running laptop and the ESP8266 to the same router,
// it should work.
#endif


//////////  setup function  //////////////

void setup() {

  delay(500);
  Serial.begin(UART_BAUD);
#if !defined(DEBUG_MODE) || DEBUG_MODE == 0  
#pragma message "ESP8266 IN NORMAL OPERATING MODE"
#pragma message "serial Rx to GPIO pin 13 and Tx to GPIO pin 15"
  pinMode(15, OUTPUT); // just to be sure...
  delay(200); // again just to be sure...
  Serial.swap(); // move UART0 pins, Rx to GPIO pin 13 and Tx to GPIO pin 15"
  delay(200);  // again just to be sure...
#endif

#if defined(MODE_AP) && MODE_AP == 1 
#pragma message "ESP8266 IN ACCESS POINT (AP) MODE"
  WiFi.mode(WIFI_AP);
  WiFi.softAPConfig(ip, ip, netmask); // ip address for the software AP 
  WiFi.softAP(ssid, pw); // ssid and password for software AP
#endif

  
#if defined(MODE_STA) && MODE_STA == 1
#pragma message "ESP8266 IN STATION (STA) MODE"
  // ESP8266 will conect to this router on which the paparazzi running laptop is also connected.
  WiFi.mode(WIFI_STA);
  WiFi.begin(ssid, pw);
  // Whait untill we are connected to the router.
  while (WiFi.status() != WL_CONNECTED) {
    delay(100);
  }
#endif

  Serial.println("Starting UDP Server");
  udp.begin(port); // start UDP server 
}

/////////// main function's while(1) loop  ////////////////
void loop() {

  // if there is udp incoming data available, read a packet
  int packet_size = udp.parsePacket();
  if(packet_size > 0) {
    remote_ip = udp.remoteIP(); // store the ip of the remote device
    remote_port = udp.remotePort(); // store the port of the remote device
    int len = udp.read(udp_buffer, BUFFER_SIZE);
#if defined(DEBUG_MODE) && DEBUG_MODE == 1
#pragma message "ESP8266 IN DEBUG MODE"
#pragma message "Packets forwarded to ttyUSB0"
    // For testing with netcat (type nc -u 192.168.4.1 4243 in a console and hit enter.)
    // Whatever you type in the console should appear in ttyUSB0 after you hit enter.
    if (len > 0){
      udp_buffer[len] = 0; // write the terminating zero or null character so udp_buffer now holds a string.
    }
    Serial.printf("Received %d bytes from %s, port %d\n", packet_size, udp.remoteIP().toString().c_str(), udp.remotePort());
    Serial.printf("UDP packet contents: %s\n", udp_buffer);
#else
    // send the received packet's payload to UART:
    Serial.write(udp_buffer, packet_size);
#endif
  } 

  // if something is received from UART0 store it to a buffer and forward it to UDP
  if(Serial.available()) {
    while(1) {
      if(Serial.available()) {
        uart_buffer[uart_idx++] = (char)Serial.read(); // read a char from UART
        if(uart_idx >= sizeof(uart_buffer)) {
           break; 
        }  
      } else {  // we finished reading the received characters.
        delay(PACKET_TIMEOUT);
        if(!Serial.available()) {
          break;
        }
      }
    }
    // send the uart received data as udp datagrams with WiFi  
    udp.beginPacket(remote_ip, 4242); // remote IP and port fro the paparazzi link agent.
    //udp.beginPacket(remote_ip, remote_port); // Use this line for general use, needs testing.
    udp.write(uart_buffer, uart_idx); // uart_idx already holds the number of characters received from UART0
    udp.endPacket();
    uart_idx = 0;  // RESET the buffer index.
  }
    
}
